@@ -5,10 +5,10 @@
 #
 # File        : apksigcopier
 # Maintainer  : Felix C. Stegerman <flx@obfusk.net>
-# Date        : 2021-03-29
+# Date        : 2021-03-30
 #
 # Copyright   : Copyright (C) 2021  Felix C. Stegerman
-# Version     : v0.2.0
+# Version     : v0.3.0
 # License     : GPLv3+
 #
 # --                                                            ; }}}1
@@ -18,6 +18,7 @@ copy/extract/patch apk signatures
 """
 
 import glob
+import logging
 import os
 import re
 import shutil
@@ -29,13 +30,14 @@ import zlib
 
 from collections import namedtuple
 
-__version__ = "0.2.0"
+__version__ = "0.3.0"
 name = "apksigcopier"
 
 SIGBLOCK, SIGOFFSET = "APKSigningBlock", "APKSigningBlockOffset"
 APKSIGNER_CMD, KEYTOOL_CMD, ZIP_CMD = "apksigner keytool zip".split()
 NOAUTOYES = NO, AUTO, YES = "no auto yes".split()
 APK_META = re.compile(r"^META-INF/[0-9A-Za-z_-]+\.(SF|RSA|MF)$")
+DATETIMEZERO = (1980, 0, 0, 0, 0, 0)
 
 ZipData = namedtuple("ZipData", "cd_offset eocd_offset cd_and_eocd".split())
 
@@ -48,6 +50,10 @@ class NoAPKSigningBlock(APKSigningBlockError):
     pass
 
 
+class ZipError(Exception):
+    pass
+
+
 # FIXME
 class ReproducibleZipInfo(zipfile.ZipInfo):
     """Reproducible ZipInfo hack."""
@@ -77,7 +83,7 @@ class APKZipInfo(ReproducibleZipInfo):
         compress_type=8,
         create_system=0,
         create_version=20,
-        date_time=(1980, 0, 0, 0, 0, 0),
+        date_time=DATETIMEZERO,
         external_attr=0,
         extract_version=20,
         flag_bits=0x800,
@@ -86,10 +92,12 @@ class APKZipInfo(ReproducibleZipInfo):
 
 # FIXME: .DSA?
 def is_meta(filename):
+    """Returns whether filename is an APK signature file."""
     return APK_META.fullmatch(filename) is not None
 
 
 def noautoyes(value):
+    """Turns False into NO, None into AUTO, and True into YES."""
     if isinstance(value, str):
         if value not in NOAUTOYES:
             raise ValueError("expected NO, AUTO, or YES")
@@ -99,6 +107,7 @@ def noautoyes(value):
 
 # FIXME
 def remove_from_zip(zipfile, meta, zip_cmd=None):
+    """Remove signature files appended by apksigner from zip file."""
     if zip_cmd and shutil.which(zip_cmd):
         files = [info.filename for info in meta]
         subprocess.run([zip_cmd, "-q", "-d", zipfile] + list(files), check=True)
@@ -146,6 +155,85 @@ def sign_with_dummy_key(output_apk, keystore, alias="dummy",
     subprocess.run(args, check=True)
 
 
+def copy_zipfile(unsigned_apk, output_apk, exclude=is_meta):
+    """Copy zip file (without signature files)."""
+    with zipfile.ZipFile(unsigned_apk, "r") as zfi:
+        with zipfile.ZipFile(output_apk, "w") as zfo:
+            for info in zfi.infolist():
+                if not exclude(info.filename):
+                    zfo.writestr(info, zfi.read(info.filename))
+
+
+# FIXME
+def copy_apk(unsigned_apk, output_apk):
+    """Copy APK (without MANIFEST.MF); returns max date_time."""
+    exclude = ["META-INF/MANIFEST.MF"]
+    with zipfile.ZipFile(unsigned_apk, "r") as zf:
+        infos = zf.infolist()
+    date_time = max(info.date_time for info in infos)
+    if not any(info.filename in exclude for info in infos):
+        shutil.copy(unsigned_apk, output_apk)
+        return date_time
+    zdata = zip_data(unsigned_apk)
+    offsets = {}
+    with open(unsigned_apk, "rb") as fhi, open(output_apk, "w+b") as fho:
+        for info in sorted(infos, key=lambda info: info.header_offset):
+            off_i = fhi.tell()
+            if info.header_offset > off_i:
+                fho.write(fhi.read(info.header_offset - off_i))
+            hdr = fhi.read(30)
+            if hdr[:4] != b"\x50\x4b\x03\x04":
+                raise ZipError("Expected local file header signature")
+            n, m = struct.unpack("<HH", hdr[26:30])
+            hdr += fhi.read(n + m)
+            if info.filename in exclude:
+                fhi.seek(info.compress_size, os.SEEK_CUR)
+            else:
+                off_o = fho.tell()
+                if info.compress_type == 0:
+                    old_off = info.header_offset + 30 + n + m
+                    new_off = off_o + 30 + n
+                    align = 4096 if info.filename.endswith(".so") else 4
+                    if len(info.extra) >= 6 and info.extra[:2] == b"\x35\xd9":
+                        dsize = int.from_bytes(info.extra[2:4], "little")
+                        if dsize > 1 and dsize <= len(info.extra) - 4:
+                            align = int.from_bytes(info.extra[4:6], "little")
+                    if old_off % align == 0 and new_off % align != 0:
+                        pad = (align - (new_off + 6) % align) % align
+                        xtr = struct.pack("<HHH", 0xd935, 2 + pad, align) + pad * b"\x00"
+                        m_b = int.to_bytes(len(xtr), 2, "little")
+                        hdr = hdr[:28] + m_b + hdr[30:30 + n] + xtr
+                offsets[id(info)] = off_o
+                fho.write(hdr)
+                size = info.compress_size
+                while size > 0:
+                    data = fhi.read(min(size, 4096))
+                    size -= len(data)
+                    fho.write(data)
+        size = zdata.cd_offset - fhi.tell()
+        while size > 0:
+            data = fhi.read(min(size, 4096))
+            size -= len(data)
+            fho.write(data)
+        cd_offset = fho.tell()
+        for info in infos:
+            hdr = fhi.read(46)
+            if hdr[:4] != b"\x50\x4b\x01\x02":
+                raise ZipError("Expected CD file header signature")
+            n, m, k = struct.unpack("<HHH", hdr[28:34])
+            hdr += fhi.read(n + m + k)
+            if info.filename not in exclude:
+                off = int.to_bytes(offsets[id(info)], 4, "little")
+                hdr = hdr[:42] + off + hdr[46:]
+                fho.write(hdr)
+        eocd_offset = fho.tell()
+        fho.write(zdata.cd_and_eocd[zdata.eocd_offset - zdata.cd_offset:])
+        fho.seek(eocd_offset + 8)
+        fho.write(struct.pack("<HHLL", len(offsets), len(offsets),
+                              eocd_offset - cd_offset, cd_offset))
+    return date_time
+
+
 def extract_meta(signed_apk):
     """Extract v1 signature metadata files from signed APK."""
     with zipfile.ZipFile(signed_apk, "r") as zf_sig:
@@ -154,25 +242,32 @@ def extract_meta(signed_apk):
                 yield info, zf_sig.read(info.filename)
 
 
-def replace_meta(extracted_meta, output_apk, zip_cmd=None):
-    """Replace v1 signature metadata in signed APK (removes v2 sig block)."""
+def replace_meta(extracted_meta, output_apk, zip_cmd=None, date_time=DATETIMEZERO):
+    """Replace v1 signature metadata in signed APK; removes v2 sig block."""
     with zipfile.ZipFile(output_apk, "r") as zf_out:
         meta = [info for info in zf_out.infolist() if is_meta(info.filename)]
-        date_time = meta[0].date_time                          # FIXME
-        remove_from_zip(output_apk, meta, zip_cmd=zip_cmd)
+        if meta:
+            date_time = meta[0].date_time                      # FIXME
+            remove_from_zip(output_apk, meta, zip_cmd=zip_cmd)
     with zipfile.ZipFile(output_apk, "a") as zf_out:
-        if "compresslevel" in zipfile.ZipFile.__doc__:         # FIXME
-            for info, data in extracted_meta:
-                zf_out.writestr(APKZipInfo(info, date_time=date_time),
-                                data, compresslevel=9)
-        else:
-            old = zipfile._get_compressor
-            zipfile._get_compressor = lambda _: zlib.compressobj(9, 8, -15)
-            try:
-                for info, data in extracted_meta:
-                    zf_out.writestr(APKZipInfo(info, date_time=date_time), data)
-            finally:
-                zipfile._get_compressor = old
+        info_data = [(APKZipInfo(info, date_time=date_time), data)
+                     for info, data in extracted_meta]
+        _write_to_zip(info_data, zf_out)
+
+
+if "compresslevel" in zipfile.ZipFile.__doc__:                 # FIXME
+    def _write_to_zip(info_data, zf_out):
+        for info, data in info_data:
+            zf_out.writestr(info, data, compresslevel=9)
+else:
+    def _write_to_zip(info_data, zf_out):
+        old = zipfile._get_compressor
+        zipfile._get_compressor = lambda _: zlib.compressobj(9, 8, -15)
+        try:
+            for info, data in info_data:
+                zf_out.writestr(info, data)
+        finally:
+            zipfile._get_compressor = old
 
 
 # https://source.android.com/security/apksigning/v2#apk-signing-block-format
@@ -257,15 +352,25 @@ def patch_apk(extracted_meta, extracted_v2_sig, unsigned_apk,
     apksigner_cmd = config.get("apksigner_cmd", APKSIGNER_CMD)
     keytool_cmd = config.get("keytool_cmd", KEYTOOL_CMD)
     zip_cmd = config.get("zip_cmd", None)
-    shutil.copy(unsigned_apk, output_apk)
-    if dummyks is None:
-        with tempfile.TemporaryDirectory() as tmpdir:
-            dummyks = os.path.join(tmpdir, "dummy-keystore")
-            gen_dummy_key(dummyks, keytool_cmd=keytool_cmd)
-            sign_with_dummy_key(output_apk, dummyks, apksigner_cmd=apksigner_cmd)
+    date_time = DATETIMEZERO
+    if not (apksigner_cmd and shutil.which(apksigner_cmd)):
+        logger = logging.getLogger(__name__)
+        if extracted_v2_sig is None:
+            logger.warning("Patching without apksigner (v1)")
+            copy_zipfile(unsigned_apk, output_apk)
+        else:
+            logger.warning("Patching without apksigner (v2)")
+            date_time = copy_apk(unsigned_apk, output_apk)
     else:
-        sign_with_dummy_key(output_apk, dummyks, apksigner_cmd=apksigner_cmd)
-    replace_meta(extracted_meta, output_apk, zip_cmd=zip_cmd)
+        shutil.copy(unsigned_apk, output_apk)
+        if dummyks is None:
+            with tempfile.TemporaryDirectory() as tmpdir:
+                dummyks = os.path.join(tmpdir, "dummy-keystore")
+                gen_dummy_key(dummyks, keytool_cmd=keytool_cmd)
+                sign_with_dummy_key(output_apk, dummyks, apksigner_cmd=apksigner_cmd)
+        else:
+            sign_with_dummy_key(output_apk, dummyks, apksigner_cmd=apksigner_cmd)
+    replace_meta(extracted_meta, output_apk, zip_cmd=zip_cmd, date_time=date_time)
     if extracted_v2_sig is not None:
         patch_v2_sig(extracted_v2_sig, output_apk)
 
@@ -395,6 +500,20 @@ def main():
     def gen_dummy(dummy_keystore, keysize):
         gen_dummy_key(dummy_keystore, keysize=keysize)
 
+    # FIXME
+    if click.__version__.startswith("7."):
+        def autocomplete_path(ctx=None, args=(), incomplete=""):  # pylint: disable=W0613
+            head, tail = os.path.split(incomplete)
+            return sorted(
+                (e.path if head else e.path[2:]) + ("/" if e.is_dir() else "")
+                for e in os.scandir(head or ".") if e.name.startswith(tail)
+            )
+
+        for command in cli.commands.values():
+            for param in command.params:
+                if isinstance(param.type, click.Path):
+                    param.autocompletion = autocomplete_path
+
     cli(prog_name=name)
 
 
